import whois
import logging
import requests

def whois_plugin(data):
    try:
        hostDict = whois.whois(data)
        hostDict = dict(hostDict)
        # change domain list from list to string
        if isinstance(hostDict.get('domain_name'), list):
            hostDict['domain_name'] = hostDict['domain_name'][0].lower()

        # add verified field
        hostDict.update({ 'verified': False })

        # search for domain in 500DB
        urlManifest = requests.get('https://raw.githubusercontent.com/Cyvore/IconDB/master/Resources/top500urls.txt').text.split('\n')
        for url in urlManifest:
            if hostDict['domain_name'] in url:
                hostDict['verified'] = True
                break

        return hostDict
    except Exception as e:
        logging.warning(e)

def run_check(chk):
    plugin_name = "Whois"
    output = whois_plugin(chk.raw)
    chk.add_plugin(plugin_name,output)
        
def describe():
    desc = """This plugin query domain/ip in whois database """
    return desc

def tags():
    # Todo: consider adding support for ips
    tags_list = ["domain"]
    return tags_list


import vt
import time
import os
import logging

VIRUS_TOTAL_KEY = os.environ['VIRUS_TOTAL_KEY']
WAIT = 4.5
MAX_TRIES = 4

def virusTotalCheck(url):
    try:
        client = vt.Client(VIRUS_TOTAL_KEY)
        analysis = client.scan_url(url)
        cur = 0
        while cur < MAX_TRIES:
          analysis = client.get_object("/analyses/{}", analysis.id)
          if analysis.status == "completed":
            return analysis.to_dict()
          cur += 1
          time.sleep(WAIT)
    except Exception as e:
        logging.info(e)
    return ""

def run_check(chk):
    plugin_name = "VirusTotal"
    output = virusTotalCheck(chk.raw)
    chk.add_plugin(plugin_name,output)
        
def describe():
    desc = """This plugin query url/ip in VirusTotal v3 database """
    return desc

def tags():
    tags_list = ["url", "domain"]
    return tags_list



from datetime import datetime
import socket
import logging
import re
import ipaddress
import urlexpander
from urllib.parse import urlparse

# MIME libraries
from eml_parser import eml_parser
import extract_msg

IPV4REGEX  = r"\b(?:[0-9]{1,3}\.){3}[0-9]{1,3}\b"
IPV6REGEX  = r"(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))"
URLREGEX = r"(?i)\b((?:https?://|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:'\".,<>?«»“”‘’]))"
EMAILREGEX = r"\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b"
BTCREG     = r"(?:bc1|[13])[a-zA-HJ-NP-Z0-9]{25,39}"
DASHREG    = r"X[1-9A-HJ-NP-Za-km-z]{33}"
LTCREG     = r"[LM3][a-km-zA-HJ-NP-Z1-9]{26,33}"
DOGEREG    = r"D{1}[5-9A-HJ-NP-U]{1}[1-9A-HJ-NP-Za-km-z]{32}"
COINS      = [BTCREG, DASHREG, LTCREG, DOGEREG]

class Plugin:
    """
    Plugin is part of check type which holds all plugins output for a check 
    """
    def __init__(self, checkID, pluginName, raw, output):
        self.checkID = checkID
        self.pluginName = pluginName
        self.raw = raw
        self.output = output
        self.timestamp = datetime.now().strftime("%m%d%Y%H%M%S")
    
    def getDict(self):
        """
        Convert plugin object into dictionary
        """
        plugin_dict = {"checkID" : self.checkID, "pluginName": self.pluginName, "raw" : self.raw,"output": self.output, "timestamp" : self.timestamp}
        return plugin_dict
      
class Check:
    """
    Check is an object to test against new plugins.
    When check is made as part of a Case object it will hold one value, url/file/crypto wallet.
    When check is self made it could hold all types of data in raw. 
    """
    def __init__(self, id, raw, tag=[]):
        self.raw = raw
        self.reputation = 0
        self.hash = ""
        self.plugins = []
        self.checkID = self.getID()
        self.id = id
        self.tags = []
        if tag and type(tag) == list:
            self.tags = tag
        self.timestemp = datetime.now().strftime("%m%d%Y%H%M%S")
    
    def getDict(self):
        """
        Convert check object into dictionary
        """
        check_dict = {"rawData" : self.raw, "reputation": self.reputation, "checkID" : self.checkID, "plugins" : [], "hash" : self.hash}
        for plg in self.plugins:
            check_dict["plugins"].append(plg.getDict())
        return check_dict
        
    def getID(self):
        """
        Create ID from the checked value: 
         - hash   (if not exist)
         - url    (if not exist)
         - wallet (if not exist)
         - empty string.
        """
        # logic TBD
        if self.hash != "":
            id = self.hash
        elif self.raw != "":
            id = self.raw
        else:
            id = ""
        return id
    
    def isEmpty(self):
        """
        Boolean funtion: return false if any of the fields is set. 
         - hash    
         - url     
         - wallet  
         - checkID 
         """
        if self.raw == "" and self.hash == "" and self.checkID == "":
            return True
        return False
    
    def add_plugin(self,pluginName, output):
        """
        Boolean function: returns true if plugin successfully added
        """
        if output == "":
            return False
        current_plugin = Plugin(self.checkID, pluginName, self.raw, output)
        self.plugins.append(current_plugin)
        return True

 
class Case:
    """
    Case is an object to investigate multiple leads from the same source. 
    checkArray will hold every lead and will only repersent one value - url/file/crypto wallet.
    """  
    def __init__(self, raw, empty=False, customID=None):
        logging.info("Initializing Case")
        self.id = customID or self.getid()
        self.checkArray = []
        self.raw = raw
        if not empty:
            self.createChecks()
        self.timestemp = datetime.now().strftime("%m%d%Y%H%M%S")
        logging.debug(f"Created case {self.id} with {self.size()} checks")
        
    def getid(self):
        """
        Create case ID from current time and host name, may changed TBD
        """
        # logic TBD
        timeStamp = datetime.now().strftime("%m%d%Y%H%M%S")
        hostPart = socket.gethostname()
        id = "%s-%s"%(timeStamp, hostPart)
        return id

    def urlAndDomainChecks(self):
        """
        Create check for every unique urls and domain in raw data
        """
        try:
            logging.info("Querying for URLs")
            urls = re.findall(URLREGEX, self.raw)
            if len(urls) > 0:        
                logging.debug("Create checks for URLs and Domains:")
                urls = [url[0] for url in urls]
                # Casting for getUniques.
                if type(urls) != list or type(urls) != tuple:
                    urls = list(urls)
                for url in self.getUniquesUrls(urls):
                    try:
                        if urlexpander.is_short(url):
                            url = urlexpander.expand(url)
                    except Exception as e:
                        logging.info(e)
                    tmpChk = Check(self.id, url,["url"])
                    self.checkArray.append(tmpChk)   
                    logging.debug(f"\t{url}") 
                    try:
                        domain = urlparse(url).netloc
                        if domain.startswith("www."):
                            domain = domain[4::]
                        tmpChk = Check(self.id, domain,["domain"])
                        self.checkArray.append(tmpChk)
                    except Exception as e:
                        logging.info(e)

            else:
                logging.warning(f"No URLs found in case.")
        except Exception as e:
            logging.info(e)
            return ""
    
    def ipChecks(self):
        """
        Create check for every unique ip in raw data
        """
        logging.info("Querying for IPs")
        ips =  re.findall(IPV4REGEX, self.raw) + re.findall(IPV6REGEX, self.raw)
        if len(ips) > 0:
            logging.debug("Create checks for URLs:")
            for cur_ip in self.getUniques(ips):
                try:
                    ip = ipaddress.ip_address(cur_ip)
                    tmpChk = Check(self.id, ip.exploded, ["ip"])
                    self.checkArray.append(tmpChk)   
                    logging.debug(f"\t{ip.exploded}") 
                except ValueError:
                    logging.debug(f'address/netmask is invalid: {cur_ip}')
    
    def emailChecks(self):
        """
        Create check for every unique email addresses in raw data
        """
        try:
            logging.info("Querying for Email addresses")
            emails_ad =  re.findall(EMAILREGEX, self.raw)
            if len(emails_ad) > 0:        
                logging.debug("Create checks for Email addresses:")

                # Casting for getUniques.
                if type(emails_ad) != list or type(emails_ad) != tuple:
                    email_ad = list(emails_ad)
                for email_ad in self.getUniques(emails_ad):
                    tmpChk = Check(self.id, email_ad, ["email"])
                    self.checkArray.append(tmpChk)   
                    logging.debug(f"\t{email_ad}") 
            else:
                logging.warning(f"No Email addresses found in case.")
        except Exception as e:
            logging.info(e)
            return ""

    def walletsCheck(self):
        """
        Create check for every unique crypto addresses in raw data
        """
        try:
            logging.info("Querying for crypto addresses")
            for coin in COINS:
                wallat_ad = re.findall(coin, self.raw)
                if len(wallat_ad) > 0:        
                    logging.debug("Create checks for crypto addresses:")
                    for cur_wallet in self.getUniques(wallat_ad):
                        tmpChk = Check(self.id, cur_wallet, ["crypto"])
                        self.checkArray.append(tmpChk)   
                        logging.debug(f"\t{cur_wallet}") 
                else:
                    logging.warning(f"No Crypto addresses found in case.")
        except Exception as e:
            logging.info(e)
            return ""

    def getUniques(self, data):
        unique_data = []
        for i in data: 
            # check if exists in unique_list or not 
            if i not in unique_data: 
                unique_data.append(i) 
        return unique_data
   
    def getUniquesUrls(self, data):
        unique_data = []
        option1, option2 = '', ''
        for i in data: 
            if not re.match(r"https?://", i):
                option1 = 'https://' + i
                option2 = 'http://' + i
            # check if exists in unique_list or not 
            if i not in unique_data and option1 not in unique_data and option2 not in unique_data: 
                unique_data.append(i) 
        return unique_data
    def size(self):
        """
        Return the amount of checks in the case
        """
        return len(self.checkArray)
    def emailFileCheck(self):
        magicNumbers = { 'eml': [bytes([0x44, 0x65, 0x6c, 0x69, 0x76, 0x65, 0x72, 0x65, 0x64]),
                                  bytes([0x52, 0x65, 0x74, 0x75, 0x72, 0x6e, 0x2d, 0x50]),
                                  bytes([0x46, 0x72, 0x6f, 0x6d]),
                                  bytes([0x58, 0x2d]),
                                  bytes([0x23, 0x21, 0x20, 0x72, 0x6e, 0x65, 0x77, 0x73]),
                                  bytes([0x46, 0x6f, 0x72, 0x77, 0x61, 0x72, 0x64, 0x20, 0x74, 0x6f]),
                                  bytes([0x46, 0x72, 0x6f, 0x6d, 0x3a]),
                                  bytes([0x4e, 0x23, 0x21, 0x20, 0x72, 0x6e, 0x65, 0x77, 0x73]),
                                  bytes([0x50, 0x69, 0x70, 0x65, 0x20, 0x74, 0x6f]),
                                  bytes([0x52, 0x65, 0x63, 0x65, 0x69, 0x76, 0x65, 0x64, 0x3a]),
                                  bytes([0x52, 0x65, 0x6c, 0x61, 0x79, 0x2d, 0x56, 0x65, 0x72, 0x73, 0x69, 0x6f, 0x6e, 0x3a]),
                                  bytes([0x52, 0x65, 0x74, 0x75, 0x72, 0x6e, 0x2d, 0x50, 0x61, 0x74, 0x68, 0x3a]),
                                  bytes([0x52, 0x65, 0x74, 0x75, 0x72, 0x6e, 0x2d, 0x70, 0x61, 0x74, 0x68, 0x3a]),
                                  bytes([0x53, 0x75, 0x62, 0x6a, 0x65, 0x63, 0x74, 0x3a, 0x20])], 
                          'msg': bytes([0xD0, 0xCF, 0x11, 0xE0, 0xA1, 0xB1, 0x1A, 0xE1]) }
        try:
            parsedMime = {}
            
            # gmail- eml
            if any (self.raw.startswith(magicNumber) for magicNumber in magicNumbers['eml']):
                ep = eml_parser.EmlParser(include_raw_body=True, include_attachment_data=True)
                parsedMime = ep.decode_email_bytes(self.raw)
                tmpChk = Check(self.id, parsedMime, ["mail"])
                self.checkArray.append(tmpChk) 
                # parsedMime = str(parsedMime.get('attachment') or '')
                parsedMime = str(parsedMime['body']) + str(parsedMime['header']['header'].get('reply-to') or [])
                
            # outlook- msg
            elif self.raw.startswith(magicNumbers['msg']):
                tmpChk = Check(self.id, parsedMime, ["mail"])
                self.checkArray.append(tmpChk)
                parsedMime = extract_msg.openMsg(self.raw)
                # parsedMime = str(parsedMime.attachments)
                parsedMime = str(parsedMime.inReplyTo) + str(parsedMime.body)
            else:
                return "Received a file that is not .eml or .msg"

            
            self.raw = parsedMime

        except Exception as e:
            logging.info(e)
            return ""

    def createChecks(self):
        """
        Create checks array from raw data, check could be either one url/file/crypto wallet.
        Changing self.checkArray. 
        """

        # First create check MUST be emailFileCheck
        logging.info("Creating Checks...")
        # try:
        #     self.emailFileCheck()
        # except Exception as e:
        #     logging.warning(e)
        try:
            self.urlAndDomainChecks()
        except Exception as e:
            logging.warning(e)
        try:
            self.ipChecks()
        except Exception as e:
            logging.warning(e)
        try:
            self.emailChecks()
        except Exception as e:
            logging.warning(e)
        try:
            self.walletsCheck()
        except Exception as e:
            logging.warning(e)
        for chk in self.checkArray:
            logging.debug(f"\t {chk.raw}")
    
    def getDict(self):
        """
        Convert case object into dictionary
        """
        case_dict = {"id" : self.id, "raw" : self.raw, "checks": [] ,"timestamp" : self.timestemp}
        for chk in self.checkArray:
            case_dict["checks"].append(chk.getDict())
        return case_dict



import json
import requests
from datetime import datetime
from datetime import timedelta
from blockcypher import get_address_overview
from blockcypher import get_address_details
from requests.exceptions import ReadTimeout

# Coin options: bitcoin, litecoin, dogecoin, dashcoin
coins = ['btc', 'ltc', 'doge', 'dash']  

# Consts
amountToCheck = 30
suspiciousAmount = amountToCheck / 2
suspiciousRange = 0 
# Needs to be tuned, If increased, more wallets would fall under the definition of suspicious transaction error

class BlackListError(Exception):
    """Raised when the address appears on a blacklist"""
    pass

class transactionError(Exception):
    """Raised when the address appears to have made suspicious transactions"""
    pass

class Verified(Exception):
    """Raised when all the validators verified the address"""
    pass

class transaction:
    """
    representation of a single transaction by date and balance in the account
    """

    def __init__(self, TranDate:datetime, balance:int):
        self.balance = balance
        self.TranDate = TranDate
    
    @property
    def date(self):
        return self.TranDate
    
    def balance(self):
        return self.balance
    
    def __lt__(self, other):
        return self.date < other.date
    
    def __ge__(self, other):
        return not self < other

class transactionArray:
    """
    Class that holds all the transactions and balance in a crypto account
    Available manipulations on the data structure: sorting, average(with a range) and two validations
    """
    
    def __init__(self) -> None:
        self.arr = []

    def add(self, newT:transaction):
        self.arr.append(newT)
    
    def sort(self):
        self.arr.sort()
    
    def setSuspiciousRange(self):
        """ tune the suspicious range to be appopriate to the wallet owner"""
        """ calculating the median, if checks that length is odd"""
        global suspiciousRange
        n = len(self.arr) - 1
        suspiciousRange = abs(self.arr[n//2].balance - self.arr[n//2 + 1].balance) // 10 if n > 1 else 0
    
    def transfer(self, index):
        """ a transfer amount is defined as difference of balance before transfer and balance after transfer"""
        return abs(self.arr[index].balance - self.arr[index + 1].balance)

    def average(self, start, amountToCheck):
        """
        average of transfers
        """
        balance = 0
        amountChecked = 0 
        
        try:
            while amountChecked is not amountToCheck:
                balance += self.transfer(start + amountChecked)
                amountChecked += 1
        except IndexError:
            pass

        return balance / amountChecked

    def suspiciousActivityByAmount(self):
        """
        check consecutive amount of transactions for suspicious behavior.
        Note that this validation is different than the validation by date because it doesn't mind dates of transactions,
        meaning, the transactions could had been done throught a decade and this validation would've noticed it.
        """

        for i in range(0, len(self.arr) - 1):
            avg = self.average(i, amountToCheck)
            amountChecked = 0
            counter = 0
            try:
                while amountChecked is not amountToCheck:
                    if abs(avg - self.transfer(i + amountChecked)) < suspiciousRange:
                        counter += 1
                    amountChecked += 1
            except IndexError:
                pass
            if counter > suspiciousAmount:
                raise transactionError
        return False

    def checkForScam(self, r, l):
        """
        called by suspiciousActivityByDate, checks for suspicious transfer routine
        """
        avg = self.average(r, l-r)
        counter = 0

        for i in range(r, l):
            if abs(self.transfer(i) - avg) < suspiciousRange:
                counter += 1
        
        # l - r - 3 is the suspicious amount of transactions, TBD about the number 3.
        if counter >= l - r - 3:
            raise transactionError

    def suspiciousActivityByDate(self):
        """
        Checking suspicious routine in transactions.
        The function sets a time limit on a routine of transactions and checks whether they were suspiciously monotonic
        which indicates a phishing scam.
        Algorithm used: window algorithm on a sorted list for achieving correct ranges in O(n) run-time
        """
        
        timeLimit = timedelta(days=amountToCheck)
        r, l = 0, 0

        while l < (len(self.arr) - 2):
            l += 1
            
            # l-r > 15 because we need at least 15 suspicious transactions to suspect a scam
            if timeLimit < self.arr[l].date - self.arr[r].date and l-r > suspiciousAmount:
                self.checkForScam(r, l - 1)

                # Resize the window to avoid repeating checks
                while timeLimit < self.arr[l].date - self.arr[r].date and r < l:   
                    r += 1
        raise Verified


def blacklistValid(wallet) -> bool:
    """
    Second validation: Appearance on a blacklist
    """

    URLBLACKLIST = f'https://api.cryptoscamdb.org/v1/check/{wallet}'

    payload={}
    headers = {}

    try:
        response = requests.get(URLBLACKLIST, headers=headers, data=payload, timeout=30) # stuck if VPN is active
    except Exception as e:
        if isinstance(e, ReadTimeout):
            return True

    info = json.loads(response.content.decode("utf-8"))
    
    if info['success'] is False:
        return True

    info = info['result']['entries'][0]['type']

    if info == 'scam':
        raise BlackListError
    return True


def transactionsValid(wallet):
    """
    Third validation: transactions 
    """

    transac = get_address_details(wallet)["txrefs"]
    db = transactionArray()

    [db.add(transaction(tx['confirmed'], tx['ref_balance'])) for tx in transac]
    db.sort()
    db.setSuspiciousRange()
    db.suspiciousActivityByAmount()
    db.suspiciousActivityByDate()


def walletVerification(wallet):
    """
    Verification of a cryptocurrency wallet using three stages:
    1. Check if the given address belongs to an existing wallet (using coinaddr)
    2. Ensure that the wallet doesn't have a history of reports and belongs to a black list (using the cryptoscamdb api)
    3. Ensure that the wallet hasn't initiated in any suspicious activity lately 
       by importing the latest transactions and ispecting their behavior (using blockcypher)
       validate(coin, wallet).valid and 
    """

    for coin in coins:
        try:
            assert get_address_overview(wallet, coin)
            assert blacklistValid(wallet)
            assert transactionsValid(wallet)

        except BlackListError:
            return 'Address is On the BlackList'
        except transactionError:
            return 'Transactions in the address are suspicious'
        except Verified:
            return 'Wallet is legitimate'
        except:
            continue
    return 'Wallet address has an issue (corrupt / non-existent)'



def run_check(chk):
    output = walletVerification(chk.raw)
    plugin_name = "cryptoWalletValidator-Plugin"
    output = "cryptoWalletValidator-Plugin check: " + chk.raw
    chk.add_plugin(plugin_name,output)
  
def describe():
    desc = """Verification of a cryptocurrency wallet"""
    return desc

def tags():
    tags_list = ["crypto"]
    return tags_list



from CyvoreOS.checkObject import Check
from CyvoreOS.checkObject import Case
import importlib
from CyvoreOS import Plugins
#from CyvoreOS import interfaces
import pkgutil
import string
import logging
#import CyvoreOS.interfaces
import urlexpander
from urllib.parse import urlparse
import re
import ipaddress

IPV4REGEX  = r"\b(?:[0-9]{1,3}\.){3}[0-9]{1,3}\b"
IPV6REGEX  = r"(([0-9a-fA-F]{1,4}:){7,7}[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,7}:|([0-9a-fA-F]{1,4}:){1,6}:[0-9a-fA-F]{1,4}|([0-9a-fA-F]{1,4}:){1,5}(:[0-9a-fA-F]{1,4}){1,2}|([0-9a-fA-F]{1,4}:){1,4}(:[0-9a-fA-F]{1,4}){1,3}|([0-9a-fA-F]{1,4}:){1,3}(:[0-9a-fA-F]{1,4}){1,4}|([0-9a-fA-F]{1,4}:){1,2}(:[0-9a-fA-F]{1,4}){1,5}|[0-9a-fA-F]{1,4}:((:[0-9a-fA-F]{1,4}){1,6})|:((:[0-9a-fA-F]{1,4}){1,7}|:)|fe80:(:[0-9a-fA-F]{0,4}){0,4}%[0-9a-zA-Z]{1,}|::(ffff(:0{1,4}){0,1}:){0,1}((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])|([0-9a-fA-F]{1,4}:){1,4}:((25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9])\.){3,3}(25[0-5]|(2[0-4]|1{0,1}[0-9]){0,1}[0-9]))"
URLREGEX = r"(?i)\b((?:https?://|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:'\".,<>?«»“”‘’]))"
EMAILREGEX = r"\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,}\b"

BTCREG     = r"(?:bc1|[13])[a-zA-HJ-NP-Z0-9]{25,39}"
DASHREG    = r"X[1-9A-HJ-NP-Za-km-z]{33}"
LTCREG     = r"[LM3][a-km-zA-HJ-NP-Z1-9]{26,33}"
DOGEREG    = r"D{1}[5-9A-HJ-NP-U]{1}[1-9A-HJ-NP-Za-km-z]{32}"
COINS      = [BTCREG, DASHREG, LTCREG, DOGEREG]

logging.basicConfig(filename='cyvore_main.log',
                    level=logging.DEBUG,
                    format='%(asctime)s | %(name)s | %(levelname)s | %(message)s')
printable = set(string.printable)

# Get absolute path for plugins
def iter_namespace(ns_pkg):
    # Specifying the second argument (prefix) to iter_modules makes the
    # returned name an absolute name instead of a relative one. This allows
    # import_module to work without having to do additional modification to
    # the name.
    return pkgutil.iter_modules(ns_pkg.__path__, ns_pkg.__name__ + ".")

# Make all Plugins dictionary 
discovered_plugins = {
    name: importlib.import_module(name)
    for finder, name, ispkg
    in iter_namespace(Plugins)
}

def exitWithLog(msg):
    if msg:
        logging.warning(msg)
    else:
        logging.info("Program finished successfully")
    logging.info("-----------------------------------------------------\n\n\n")
    exit()


def strings(filename, min=4):
    with open(filename, errors="ignore") as f:  # Python 3.x
        result = ""
        for c in f.read():
            if c in string.printable:
                result += c
                continue
            if len(result) >= min:
                yield result
            result = ""
        if len(result) >= min:  # catch result at EOF
            yield result


def listplugins():
    logging.info("Running ls option")
    for plugin in discovered_plugins:
        current_plugin = importlib.import_module(plugin)
        output = current_plugin.describe()
        logging.info("\t" + plugin, "-", output)
    exitWithLog("")


"""
# Only for windows developers 
def install():
    logging.info("Running install option")
    path = "\"C:\\Python39\\python.exe\" \"" + __file__ + "\" -uf \"%1\""
    logging.debug(f"Path in registry:\t{path}")
    # interfaces.right_click_install.define_action_on("*", "CheckForPhish", path, title="Run Phish Check")
    logging.info("Installed right click addon")
    exitWithLog("")
"""


def urlFromFileCommand(args):
    testCase = Case(str(list(strings(args.urlFromFIle)))) 
    logging.info(f"Check file:\n\t{args.urlFromFIle}")
    logging.info(f"Create case:\n\t{testCase.id}")
    return testCase


def scanstring(testdata, tags=False):
    testCase = Case(testdata)
    logging.info(f"Check string:\n\t{testdata}")
    logging.info(f"Create case:\n\t{testCase.id}")
    for plugin in discovered_plugins:
        for chk in testCase.checkArray:
            current_plugin = importlib.import_module(plugin)
            if tags:
                if any(tag in chk.tags for tag in current_plugin.tags()):
                    current_plugin.run_check(chk)
                else:
                    plugin_name = str(plugin).split(".")[-1].strip("_plugin")
                    logging.info(f"Skip plugin {plugin_name} because of tags mismatch: {current_plugin.tags()}")
            else:
                current_plugin.run_check(chk)
    return testCase


def runPlugins(testdata, plugins_list):
    testCase = Case(testdata)
    for plugin in discovered_plugins:
        if plugins_list:
            plugin_name = str(plugin).split(".")[1].strip("_plugin")
            if plugin_name not in plugins_list:
                logging.info(f"Skipping {plugin_name} - plugin flag is on.")
                continue
        elif "debug" in str(plugin).split(".")[1].strip("_plugin"):
            continue
        logging.debug(f"Running plugin: {str(plugin)}")
        for chk in testCase.checkArray:
            
            current_plugin = importlib.import_module(plugin)
            current_plugin.run_check(chk)


def run_plugins_for_existing_case(case, tags=[]):
    logging.info(f"Working on existing case:\n\t{case.id}")
    for plugin in discovered_plugins:
        for chk in case.checkArray:
            current_plugin = importlib.import_module(plugin)
            if tags:
                if any(tag in chk.tags for tag in current_plugin.tags()):
                    current_plugin.run_check(chk)
                else:
                    plugin_name = str(plugin).split(".")[-1].strip("_plugin")
                    logging.info(f"Skip plugin {plugin_name} because of tags mismatch: {current_plugin.tags()}")
            else:
                current_plugin.run_check(chk)
    return case


def urlAndDomainChecks(case):
    """
    Create check for every unique urls and domain in raw data
    """
    logging.info(f"Checking Url: {case.raw}")
    if re.match(URLREGEX, case.raw):
        try:
            logging.info(f"Check if {case.raw} is shortend")
            if urlexpander.is_short(case.raw):
                url = urlexpander.expand(case.raw)
                logging.info(f"{case.raw} is shortend, got - {url}")

                # Add destination url check to case
                case.checkArray.append(Check(case.id, url, ["url"]))
            else:

                # Add url check to case
                case.checkArray.append(Check(case.id, case.raw, ["url"]))
        except Exception as e:
            logging.info(f"Error with urlexpander.expand('{case.raw}')")
            logging.info(e)

        try:
            logging.info(f"Check for {case.raw} domain")
            domain = urlparse(case.raw).netloc
            if domain.startswith("www."):
                domain = domain[4::]
            logging.info(f"in {case.raw} found domain {domain}")

            # Add domain check to case
            case.checkArray.append(Check(case.id, domain, ["domain"]))
        except Exception as e:
            logging.info(f"error with urlparse('{case.raw}').netloc")
            logging.info(e)
        return True
    else:
        logging.warning(f"No URLs found in case.")
        return False


def ipChecks(case):
    """
    Create check for every unique ip in raw data
    """
    logging.info("Querying for IPs")
    ips = re.findall(IPV4REGEX, case.raw) + re.findall(IPV6REGEX, case.raw)
    if len(ips) > 0:
        logging.debug("Create checks for IPs:")
        for cur_ip in case.getUniques(ips):
            try:
                ip = ipaddress.ip_address(cur_ip)
                case.checkArray.append(Check(case.id, ip.exploded, ["ip"]))
                logging.debug(f"\t{ip.exploded}")
                return True
            except ValueError:
                logging.debug(f'address/netmask is invalid: {cur_ip}')
    else:
        logging.warning(f"No IPs found in case.")
        return False


def scanUrl(url, tags=False):
    testCase = Case(url.strip(), empty=True)
    logging.info(f"Check Url:\n\t{url}")
    logging.info(f"Create case:\n\t{testCase.id}")
    if not urlAndDomainChecks(testCase):
        logging.info("Not found url or domain, checking for IPs")
        if not ipChecks(testCase):
            logging.info("Not found any IPs, return empty case.")
            return testCase
        
    for plugin in discovered_plugins:
        for chk in testCase.checkArray:
            current_plugin = importlib.import_module(plugin)
            if tags:
                if any(tag in chk.tags for tag in current_plugin.tags()):
                    current_plugin.run_check(chk)
                else:
                    plugin_name = str(plugin).split(".")[-1].strip("_plugin")
                    logging.info(f"Skip plugin {plugin_name} because of tags mismatch: {current_plugin.tags()}")
            else:
                current_plugin.run_check(chk)
    return testCase


def walletsCheck(case):
    """
    Create check for every unique crypto addresses in raw data
    """
    try:
        logging.info("Querying for crypto addresses")
        for coin in COINS:
            wallet_ad = re.findall(coin, case.raw)
            if len(wallet_ad) > 0:
                logging.debug("Create checks for crypto addresses:")
                for cur_wallet in case.getUniques(wallet_ad):
                    case.checkArray.append(Check(case.id, cur_wallet, ["crypto"]))
                    logging.debug(f"\t{cur_wallet}")
                    return True
            else:
                logging.warning(f"No Crypto addresses found in case.")
                return False
    except Exception as e:
        logging.info(e)
        return False


def scan_one_string(data, tags=False):
    testCase = Case(data.strip(), empty=True)
    logging.info(f"Check :\n\t{data}")
    logging.info(f"Create case:\n\t{testCase.id}")
    if not urlAndDomainChecks(testCase) or not ipChecks(testCase) or not walletsCheck(testCase):
        return ""
    for plugin in discovered_plugins:
        for chk in testCase.checkArray:
            current_plugin = importlib.import_module(plugin)
            if tags:
                if any(tag in chk.tags for tag in current_plugin.tags()):
                    current_plugin.run_check(chk)
                else:
                    plugin_name = str(plugin).split(".")[-1].strip("_plugin")
                    logging.info(f"Skip plugin {plugin_name} because of tags mismatch: {current_plugin.tags()}")
            else:
                current_plugin.run_check(chk)
    return testCase


def process(stream):
    found_str = ""
    while True:
        data = stream.read(1024*4)
        if not data:
            print("Not data")
            break
        for char in data:
            if char in printable:
                found_str += char
            elif len(found_str) >= 4:
                yield found_str
                found_str = ""
            else:
                found_str = ""



from setuptools import setup, find_packages
import os

FOLDERNAME = "CyvoreOS"

setup(
    name='CyvoreOS',
    #packages=['CyvoreOS', os.path.join(FOLDERNAME, "Plugins"), os.path.join(FOLDERNAME, "Output"), os.path.join(FOLDERNAME, "interfaces")],
    packages=find_packages(),
    package_data={"": [os.path.join(FOLDERNAME, "Resources", "top500urls.txt")]},
    include_package_data=True,
    version='0.1.7.6.4',
    license='MIT',  #https://help.github.com/articles/licensing-a-repository
    description='Next-Gen email threat prevention',
    author='Cyvore',
    author_email='info@cyvore.com',
    url='https://cyvore.com',  # Our website Link
    download_url='https://github.com/cyvore/CyvoreOS/archive/refs/tags/v_0.1.7.6.4.tar.gz',
    keywords=['Anti-Phishing', 'Email-Scanner', 'Optical-Recognition'],
    install_requires=[
        'levenshtein',
        'requests',
        'colorama',
        'future',
        'urlextract',
        'blockcypher',
        'urlexpander',
        'python-whois',
        'eml_parser',
        'python-magic',
        'extract_msg',
        "vt-py"
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3'
    ],
)



def run_check(chk):
    plugin_name = "Template-Plugin"
    output = "example for output"
    chk.add_plugin(plugin_name,output)
        
def describe():
    desc = """Description of PluginName in a sentence """
    return desc

def tags():
    tags_list = ["url", "file", "ip", "hash", "email", "domain", "crypto", "username"]
    return tags_list